/*
Very simple GLUT program using OSL libraries, demonstrating
the use of GL_ARB_vertex_program and GL_ARB_fragment_program.

Orion Sky Lawlor, olawlor@acm.org, 2005/1/26 (Public Domain)
*/
#include "ogl/ext.h"
#include "ogl/ext.cpp"
#include "ogl/glew.c"
#include "osl/viewpoint.h" /* 3D viewpoint utility routines */
#include "osl/viewpoint.cpp" /* include implementation code, for easier linking */
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

/*  This is an array of keyboard flags, indexed by key.
  Initially 0; corresponding keyboard press toggles to 1.
  For example, the x key controls oglToggles['x'].  */
int oglToggles[256]; 

void checkErr(const char *where) {
	GLenum e=glGetError();
	if (e!=0) { /* error handling-- important for debugging! */
		const GLubyte *errString = gluErrorString(e);
		printf("Error in %s\nError: %s\n",where,errString);
		exit(1);
	}
}

/*********** OpenGL Camera Definition ********/
osl::Axes3d cameraDir; /* Camera axes: Z points to camera, X and Y on screen */
double cameraDist=5.0; /* Distance to camera from cameraTarget */
osl::Vector3d cameraTarget(0,0,0); /* point camera is looking at, world coords */
double cameraAspect=1.0; /* aspect ratio-- width over height */

/* Call this whenever any camera parameters change.
   Updates the MODELVIEW matrix with the current camera location. */
void cameraChanged(void) {
	/* GL_PROJECTION contains the world->clip transform */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /* because gluPerspective combines with the current matrix */
	double z_near=0.1,z_far=1.0e3; /* z buffer range */
	gluPerspective(70.0, cameraAspect, z_near,z_far);
	
	/* Compute the location of the camera, world coords */
	osl::Vector3d cameraLocation=cameraTarget+cameraDist*cameraDir.getZ(); 
	gluLookAt(cameraLocation.x,cameraLocation.y,cameraLocation.z,
		cameraTarget.x,cameraTarget.y,cameraTarget.z,
		cameraDir.getY().x,cameraDir.getY().y,cameraDir.getY().z);
	
	/* copy camera location into vertex shader attribute */
	glProgramEnvParameter4dvARB(GL_VERTEX_PROGRAM_ARB, 0, cameraLocation);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); /* because gluLookAt adds to current matrix */
	
	/* After changing the modelview matrix, you've got to set up the 
	OpenGL light positions again, because light positions are *specified*
	in model coordinates, but OpenGL *keeps* them in eye coordinates.  
	Might as well just set up all lighting here:
	*/
	
/* Set up OpenGL lighting (*lots* of stuff to set up) */
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0); glDisable(GL_LIGHT1);
	osl::Vector3d l(0.1,1.0,-0.3); l=l.dir();
	float lightDir[4];
	lightDir[0]=l.x; lightDir[1]=l.y; lightDir[2]=l.z; lightDir[3]=0.0;
	glLightfv(GL_LIGHT0, GL_POSITION, lightDir);

	GLfloat ones[] = {1.0f, 1.0f, 1.0f, 0.0f};
	GLfloat zeros[] = {0.0f, 0.0f, 0.0f, 0.0f};
	GLfloat specularity[] = {50.0f};
	if (oglToggles['s']) specularity[0]=100.0;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,   ones);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,   ones);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,  zeros);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS,  specularity);

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	// glLightModel(GL_LIGHT_MODEL_COLOR_CONTROL,GL_SEPARATE_SPECULAR_COLOR);
	glEnable(GL_NORMALIZE);
	
	
	glutPostRedisplay(); /* ask GLUT to redraw our display */
}

void reshapeFn(int w, int h) /* Called by GLUT when window is resized */
{
	/* GL_PROJECTION matrix contains only the camera parameters */
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	cameraAspect=w/(double)h;
	cameraChanged();
}

/************ Crappy Keyboard User Interface **************/
void keyboardFn(unsigned char key, int x, int y) /* Called by GLUT when a key is pressed */
{
	switch (key) {
	case 27: case 'x': case 'q': /*  Escape or quit key  */
		exit(0);
		break;
	case '+': case '=': cameraDist*=0.94; break; /* zoom in */
	case '-': case '_': cameraDist*=1.05; break; /* zoom out */
	default: /* all other keys just toggle the corresponding "keys" entry */
		oglToggles[key]=!oglToggles[key]; 
		break;
	}
	cameraChanged(); /* ask GLUT to redraw our display */
}

/************ Crappy Mouse User Interface **************/

int lastX,lastY;
enum {mouseNone=0,mouseSpin,mouseZoom} mouseState;

void motionFn(int x, int y)
{
	double s=0.005; /* spin rate: radians of rotation per pixel of mouse movement */
	double z=0.01; /* zoom rate: zoom scale factor per pixel */
	int relX=x-lastX, relY=y-lastY;
	switch (mouseState) {
	case mouseSpin: cameraDir.nudge(relX*s,-relY*s); cameraChanged(); break;
	case mouseZoom: cameraDist*=1.0+z*relY; cameraChanged(); break;
	case mouseNone: break; /* no buttons down-- don't do anything */
	}
	lastX=x; lastY=y;
}

void mouseFn(int button, int state, int x, int y)
{
	lastX=x; lastY=y;
	if (state==GLUT_DOWN) {mouseState=(button==0)?mouseSpin:mouseZoom;}
	if (state==GLUT_UP) {mouseState=mouseNone;}
}

/************* 3D Noise Texture ************/
void uploadNoise(void) {
	const int sx=8,sy=8,sz=8;
	unsigned char noise[4*sx*sy*sz];
	for (int z=0;z<sz;z++)
	for (int y=0;y<sy;y++)
	for (int x=0;x<sx;x++) {
		unsigned char *n=&noise[4*(x+sx*(y+sy*z))];
		n[0]=0xff&rand(); // RGB all pure noise
		n[1]=0xff&rand();
		n[2]=0xff&rand();
		n[3]=0xff; // alpha is 1
	}
	glTexImage3D(GL_TEXTURE_3D,0,GL_RGBA8,
		sx,sy,sz,0,
		GL_RGBA,GL_UNSIGNED_BYTE,
		noise);
	checkErr("upload 3d tex");
	
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	checkErr("linear 3d tex");
	
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_REPEAT); 
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_REPEAT); 
	checkErr("repeat 3d tex");
}


/************ Main Drawing Function **************/

/**
  Program to run on each vertex or pixel fragment.
*/
class oglProgram {
	int target;// Program target, like GL_FRAGMENT_PROGRAM_ARB
	GLuint progName; // OpenGL handle to program
public:
	oglProgram(int target_) :target(target_) 
		{glGenProgramsARB(1, &progName);bind();}
	
	/// Make this the current OpenGL program.
	void bind(void) {
		glBindProgramARB(target, progName);
	}
	
	// Replace the current program by the one in this file.
	//  Prints out an error message if the program can't be parsed.
	void read(const char *fName);
	
	// Replace the current program by this one.
	//  Prints out an error message if the program can't be parsed.
	void set(const char *programText);
	
	~oglProgram() {glDeleteProgramsARB(1,&progName);}
};

void oglProgram::read(const char *fName) {
	
	// Read program from file
	int len=65536;
	char *s=new char[len];
	FILE *f=fopen(fName,"r");
	len=fread(s,1,len,f); // reset len to file length
	s[len]=0; // null terminate string
	fclose(f);
	
	// Upload program to graphics card
	set(s);
	
	delete[] s;
}
	
void oglProgram::set(const char *programText) {
	bind();
	glProgramStringARB(target, GL_PROGRAM_FORMAT_ASCII_ARB, 
		strlen(programText), programText);
	checkErr(programText);
}


void displayFn(void) /* Called by GLUT to show every frame */
{
/* Do basic OpenGL initialization */
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND); /* alpha blending */
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glShadeModel (GL_SMOOTH); /* gauroud (not flat) shading */
	glClearColor (0.1, 0.2, 0.3, 0.0); /* sets background color (RGBA) */

/* Main redraw */
	glClear(GL_COLOR_BUFFER_BIT+GL_DEPTH_BUFFER_BIT); /* clears to background color */
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix(); /* keeps the "translate" below from accumulating frame-to-frame */
	

/** Texture coordinate generation */
	for (int axis=0;axis<3;axis++) {
		double texGen[4]={0,0,0,0};
		texGen[axis]=1.0;
		glTexGeni(GL_S+axis,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
		glEnable(GL_TEXTURE_GEN_S+axis);
		glTexGendv(GL_S+axis,GL_OBJECT_PLANE,texGen);
	}
	
	// glEnable(GL_TEXTURE_3D);
	oglProgram vert(GL_VERTEX_PROGRAM_ARB);   vert.read("vertex.txt");
	oglProgram frag(GL_FRAGMENT_PROGRAM_ARB); frag.read("fragment.txt");
	
	// Draw stuff using fragment & vertex programs
	if (!oglToggles['f']) {
		glEnable(GL_FRAGMENT_PROGRAM_ARB);
		glEnable(GL_VERTEX_PROGRAM_ARB);
	}
	glColor3f(1,1,0); /* big yellow objects */
	
	for (double ang=0;ang<360;ang+=0.25*360) {
		glPushMatrix();
		glRotated(ang,0,0,1);
		glTranslated(0.7,0,0);
		if (oglToggles['s']) glutSolidSphere(0.4,40,60);
		else glutSolidTeapot(0.5);
		glPopMatrix();
	}
	
	/* Mark X, Y, and Z axes with little spheres */
	for (int axis=0;axis<3;axis++) {
		oglVector3d color(0.0), loc(0.0);
		color[axis]=1.0; loc[axis]=2.0;
		
		glPushMatrix();
		glColor3dv((const double *)color);
		glTranslated(loc.x,loc.y,loc.z); 
		glutSolidSphere(0.3,20,10);
		glPopMatrix();
	}
	
	glDisable(GL_FRAGMENT_PROGRAM_ARB);
	glDisable(GL_VERTEX_PROGRAM_ARB);
	
	glPopMatrix();
	glutSwapBuffers(); /* displays rendered image */
}


/******** Program Initialization *********/
int main(int argc, char** argv)
{
/* Do GLUT initialization */
	glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA);
	glutInitWindowSize (1000, 700);
	glutCreateWindow (argv[0]); /* could also do glutEnterGameMode for fullscreen */

/* Set up OpenGL Extensions */
	oglSetupExt();
	if (!GLEW_ARB_fragment_program) {
		printf("You need programmable shading to run this program!\n");
		exit(1);
	}
	uploadNoise();

/* Set up GLUT GUI routines */
	glutDisplayFunc (displayFn);
	glutReshapeFunc (reshapeFn);
	glutKeyboardFunc (keyboardFn);
	glutMouseFunc(mouseFn);
	glutMotionFunc(motionFn);
	
	/* Run the user interface until exit.  This call never returns. */
	glutMainLoop();
	return 0;
}
